private variable MODULE_NAME = "mpfp";
prepend_to_slang_load_path (".");
set_import_module_path (".:"+get_import_module_path ());
require (MODULE_NAME);

private variable Failed = 0;
private define failed ()
{
   variable args = __pop_list (_NARGS);
   () = fprintf (stderr, __push_list (args));
   () = fputs ("\n", stderr);
   Failed++;
   throw MathError;
}

private define test_class ()
{
   variable a = mpfr(1.0);
   if (typeof (a) != MPFR_Type)
     failed ("mpfr did not produce an MPFR_Type");
   
   a = MPFR_Type[10];
   if (_typeof(a) != MPFR_Type)
     failed ("Unable to create an array of MPFR_Type objects");
   
   a[0] = 2.0;
   if (typeof (a[0]) != MPFR_Type)
     failed ("An MPFR_Type array element is not an MPFR_Type");
}

private define test_unary ()
{
   variable a = mpfr(1);
   if (-a != mpfr(-1))
     failed ("unary chs operation failed");
   if (abs(-a) != a)
     failed ("unary abs operation failed");
}

private define is_equal (a, b)
{
   if ((length(a) == 0) && (length(b) == 0))
     return 1;

   variable diff = abs(a-b);
   variable abs_a = abs(a);
   variable abs_b = abs(b);
   variable res = all ((isnan(a)==isnan(b))
		       and (isnan(a) or (a==b) 
			    or (diff < 1e-15)
			    or ((abs_a > abs_b) and (diff < abs_a*1e-10))
			    or (abs(a-b) < abs_b*1e-10)));
   if (res) return res;
   %print (abs(a-b));
   return 0;
   return ((isnan(a)==isnan(b))
	   && (isnan(a) || (a==b) || (abs(a-b)<1e-15)));
   % Note; the a==b is for Inf.
}

private define test_binary_internal (x, y)
{
   variable a = mpfr(x);
   variable b = mpfr(y);
   variable z, c, ok;

   z = x + y; c = a + b;
   ifnot (is_equal(z,c))
     failed ("%S + %S operation failed", _typeof(a), _typeof(b));
   z = x + y; c = a + y;
   ifnot (is_equal(z,c))
     failed ("%S + %S operation failed", _typeof(a), _typeof(y));
   z = x + y; z = x + b;
   ifnot (is_equal(z,c))
     failed ("%S + %S operation failed", _typeof(x), _typeof(b));

   z = x - y; c = a - b;
   ifnot (is_equal(z,c))
     failed ("%S - %S operation failed", _typeof(a), _typeof(b));
   z = x - y; c = a - y;
   ifnot (is_equal(z,c))
     failed ("%S - %S operation failed", _typeof(a), _typeof(y));
   z = x - y; z = x - b;
   ifnot (is_equal(z,c))
     failed ("%S - %S operation failed", _typeof(x), _typeof(b));

   z = x * y; c = a * b;
   ifnot (is_equal(z,c))
     failed ("%S * %S operation failed", _typeof(a), _typeof(b));
   z = x * y; c = a * y;
   ifnot (is_equal(z,c))
     failed ("%S * %S operation failed", _typeof(a), _typeof(y));
   z = x * y; z = x * b;
   ifnot (is_equal(z,c))
     failed ("%S * %S operation failed", _typeof(x), _typeof(b));

   z = x / y; c = a / b;
   ifnot (is_equal(z,c))
     failed ("%S / %S operation failed", _typeof(a), _typeof(b));
   z = x / y; c = a / y;
   ifnot (is_equal(z,c))
     failed ("%S / %S operation failed", _typeof(a), _typeof(y));
   z = x / y; z = x / b;
   ifnot (is_equal(z,c))
     failed ("%S / %S operation failed", _typeof(x), _typeof(b));

   variable x_mod_y = x mod y;
   variable i = wherenot (((_typeof (x) == ULong_Type) or (_typeof(x) == UInt_Type))
			  and (y <= 0));

   ifnot (is_equal((x_mod_y[i]), (a mod b)[i]))
     failed ("%S mod %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x_mod_y[i]), (a mod y)[i]))
     failed ("%S mod %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x_mod_y[i]), (x mod b)[i]))
     failed ("%S mod %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x == y), (a == b)))
     failed ("%S == %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x == y), (a == y)))
     failed ("%S == %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x == y), (x == b)))
     failed ("%S == %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x != y), (a != b)))
     failed ("%S != %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x != y), (a != y)))
     failed ("%S != %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x != y), (x != b)))
     failed ("%S != %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x < y), (a < b)))
     failed ("%S < %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x < y), (a < y)))
     failed ("%S < %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x < y), (x < b)))
     failed ("%S < %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x <= y), (a <= b)))
     failed ("%S <= %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x <= y), (a <= y)))
     failed ("%S <= %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x <= y), (x <= b)))
     failed ("%S <= %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x > y), (a > b)))
     failed ("%S > %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x > y), (a > y)))
     failed ("%S > %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x > y), (x > b)))
     failed ("%S > %S operation failed", _typeof(x), _typeof(b));

   ifnot (is_equal((x >= y), (a >= b)))
     failed ("%S >= %S operation failed", _typeof(a), _typeof(b));
   ifnot (is_equal((x >= y), (a >= y)))
     failed ("%S >= %S operation failed", _typeof(a), _typeof(y));
   ifnot (is_equal((x >= y), (x >= b)))
     failed ("%S >= %S operation failed", _typeof(x), _typeof(b));
}

private define test_binary ()
{
   variable types = [Char_Type, UChar_Type, Short_Type, UShort_Type,
		     Int_Type, UInt_Type, Long_Type, ULong_Type,
		     Float_Type, Double_Type];
   
   variable t, x, y;
   foreach t (types)
     {
	test_binary_internal (3.0, typecast (7, t));
	test_binary_internal (typecast (3, t), 7.0);
     }
   foreach x ({7.0, _NaN, _Inf, -_Inf,})
     {
	test_binary_internal (_NaN, x);
	test_binary_internal (_Inf, x);
	test_binary_internal (x, _NaN);
	test_binary_internal (x, _Inf);
     }

   variable grid = [-5:5:0.5];
   foreach t (types)
     {
	foreach x (grid)
	  {
	     % These tests only work if a double can fully represent
	     % the object.
	     y = typecast(grid, t);
	     variable f = typecast(y, Double_Type);
	     y = y[where ((y+1 == f+1.0) and (f+1.0 != f))];
	     test_binary_internal (x, y);
	     test_binary_internal (y, x);
	  }
     }
}


private define test_math ()
{
   variable dfuns =
     {
	&sin, &cos, &tan,
	&sinh, &cosh, &tanh,
	&asin, &acos, &atan,
	&asinh, &acosh, &atanh,
	&exp, &log, &log10, &sqrt,
	&expm1, &log1p,
     };
   variable ifuns = 
     {
	&isinf, &isnan,
     };

   variable f, x;
   foreach x ({0.5, 1.0, -1.0, -0.5, _NaN, _Inf, -_Inf})
     {
	foreach f (dfuns)
	  {
	     variable a = (@f)(x);
	     variable b = (@f)(mpfr(x));
	     ifnot (is_equal(a,b))
	       {
		  failed ("%S(%S) produced %S, expected %S", 
			  f, x, b, a);
	       }
	  }
	foreach f (ifuns)
	  {
	     a = (@f)(x);
	     b = (@f)(mpfr(x));
	     ifnot ((a != 0) == (b != 0))
	       {
		  failed ("%S(%S) produced %S, expected %S", 
			  f, x, b, a);
	       }
	  }
     }
}

define slsh_main ()
{
   mpfr_set_default_prec (200);
   () = fprintf (stdout, "Testing mpfr..."); () = fflush(stdout);
   test_class ();
   test_unary ();
   test_binary ();
   test_math ();
   
   if (Failed == 0)
     () = fprintf (stdout, "Ok\n");
   else
     () = fprintf (stderr, "Failed with %d errors\n", Failed);
}


